<!doctype html>
<html>
  <head>
    <title>{{title}}</title>
    <meta charset="utf-8">
    <script type="module" src="{{rapidoc_url}}"></script>
  </head>
  <body>
    <rapi-doc
      spec-url = "{{ url_for('api-docs.openapi_json') }}"
      {% for key, val in rapidoc_config.items() -%}
        {{key}} = "{{val}}"
      {% endfor -%}
    > </rapi-doc>
  </body>
</html>
